/*--------------------------------------------------------------------------

@sinclair/typebox/type

The MIT License (MIT)

Copyright (c) 2017-2023 Haydn Paterson (sinclair) <haydn.developer@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

---------------------------------------------------------------------------*/
import { Object } from '../object/index.mjs';
import { Never } from '../never/index.mjs';
import { Union } from '../union/index.mjs';
import { IsTemplateLiteralFinite, TemplateLiteralParseExact } from '../template-literal/index.mjs';
import { PatternStringExact, PatternNumberExact } from '../patterns/index.mjs';
import { IndexPropertyKeys } from '../indexed/index.mjs';
import { Kind, Hint } from '../symbols/index.mjs';
import { CloneType } from '../clone/type.mjs';
import { IsUndefined } from '../guard/value.mjs';
// ------------------------------------------------------------------
// TypeGuard
// ------------------------------------------------------------------
// prettier-ignore
import { TInteger as IsIntegerType, TLiteral as IsLiteralType, TNumber as IsNumberType, TString as IsStringType, TTemplateLiteral as IsTemplateLiteralType, TUnion as IsUnionType, } from '../guard/type.mjs';
// ------------------------------------------------------------------
// FromPattern
// ------------------------------------------------------------------
// prettier-ignore
function FromPattern(pattern, T) {
    return { [Kind]: 'Record', type: 'object', patternProperties: { [pattern]: CloneType(T) } };
}
// ------------------------------------------------------------------
// FromKeys
// ------------------------------------------------------------------
// prettier-ignore
function FromKeys(keys, T) {
    const properties = keys.reduce((acc, key) => ({ ...acc, [key]: CloneType(T) }), {});
    return Object(properties, { [Hint]: 'Record' });
}
// prettier-ignore
function FromTemplateLiteralKey(K, T) {
    const expression = TemplateLiteralParseExact(K.pattern);
    return (IsTemplateLiteralFinite(expression)
        ? FromKeys(IndexPropertyKeys(K), T)
        : FromPattern(K.pattern, T));
}
// prettier-ignore
function FromUnionKey(K, T) {
    return FromKeys(IndexPropertyKeys(Union(K)), T);
}
// prettier-ignore
function FromLiteralKey(K, T) {
    return FromKeys([K.toString()], T);
}
// prettier-ignore
function FromStringKey(K, T) {
    const pattern = IsUndefined(K.pattern) ? PatternStringExact : K.pattern;
    return FromPattern(pattern, T);
}
// prettier-ignore
function FromIntegerKey(_, T) {
    return FromPattern(PatternNumberExact, T);
}
// prettier-ignore
function FromNumberKey(_, T) {
    return FromPattern(PatternNumberExact, T);
}
// prettier-ignore
function RecordResolve(K, T) {
    return (IsUnionType(K) ? FromUnionKey(K.anyOf, T) :
        IsTemplateLiteralType(K) ? FromTemplateLiteralKey(K, T) :
            IsLiteralType(K) ? FromLiteralKey(K.const, T) :
                IsStringType(K) ? FromStringKey(K, T) :
                    IsIntegerType(K) ? FromIntegerKey(K, T) :
                        IsNumberType(K) ? FromNumberKey(K, T) :
                            Never());
}
// prettier-ignore
/** `[Json]` Creates a Record type */
export function Record(K, T, options = {}) {
    return CloneType(RecordResolve(K, T), options);
}
