/*--------------------------------------------------------------------------

@sinclair/typebox/value

The MIT License (MIT)

Copyright (c) 2017-2023 Haydn Paterson (sinclair) <haydn.developer@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

---------------------------------------------------------------------------*/
import { TTransform as IsTransformType, TSchema as IsSchemaType } from '../../type/guard/type.mjs';
import { Kind, TransformKind } from '../../type/symbols/index.mjs';
import { IsPlainObject, IsArray, IsValueType } from '../guard/index.mjs';
import { KeyOfPropertyKeys } from '../../type/keyof/index.mjs';
import { Index } from '../../type/indexed/index.mjs';
import { Deref } from '../deref/index.mjs';
import { Check } from '../check/index.mjs';
// ------------------------------------------------------------------
// Errors
// ------------------------------------------------------------------
export class TransformEncodeCheckError extends Error {
    schema;
    value;
    error;
    constructor(schema, value, error) {
        super(`Unable to encode due to invalid value`);
        this.schema = schema;
        this.value = value;
        this.error = error;
    }
}
export class TransformEncodeError extends Error {
    schema;
    value;
    constructor(schema, value, error) {
        super(`${error instanceof Error ? error.message : 'Unknown error'}`);
        this.schema = schema;
        this.value = value;
    }
}
// ------------------------------------------------------------------
// Encode
// ------------------------------------------------------------------
// prettier-ignore
function Default(schema, value) {
    try {
        return IsTransformType(schema) ? schema[TransformKind].Encode(value) : value;
    }
    catch (error) {
        throw new TransformEncodeError(schema, value, error);
    }
}
// prettier-ignore
function TArray(schema, references, value) {
    const defaulted = Default(schema, value);
    return IsArray(defaulted)
        ? defaulted.map((value) => Visit(schema.items, references, value))
        : defaulted;
}
// prettier-ignore
function TIntersect(schema, references, value) {
    const defaulted = Default(schema, value);
    if (!IsPlainObject(value) || IsValueType(value))
        return defaulted;
    const knownKeys = KeyOfPropertyKeys(schema);
    const knownProperties = knownKeys.reduce((value, key) => {
        return key in defaulted
            ? { ...value, [key]: Visit(Index(schema, [key]), references, value[key]) }
            : value;
    }, defaulted);
    if (!IsTransformType(schema.unevaluatedProperties)) {
        return Default(schema, knownProperties);
    }
    const unknownKeys = Object.getOwnPropertyNames(knownProperties);
    const unevaluatedProperties = schema.unevaluatedProperties;
    return unknownKeys.reduce((value, key) => {
        return !knownKeys.includes(key)
            ? { ...value, [key]: Default(unevaluatedProperties, value[key]) }
            : value;
    }, knownProperties);
}
// prettier-ignore
function TNot(schema, references, value) {
    return Default(schema.not, Default(schema, value));
}
// prettier-ignore
function TObject(schema, references, value) {
    const defaulted = Default(schema, value);
    if (!IsPlainObject(value))
        return defaulted;
    const knownKeys = KeyOfPropertyKeys(schema);
    const knownProperties = knownKeys.reduce((value, key) => {
        return key in value
            ? { ...value, [key]: Visit(schema.properties[key], references, value[key]) }
            : value;
    }, defaulted);
    if (!IsSchemaType(schema.additionalProperties)) {
        return knownProperties;
    }
    const unknownKeys = Object.getOwnPropertyNames(knownProperties);
    const additionalProperties = schema.additionalProperties;
    return unknownKeys.reduce((value, key) => {
        return !knownKeys.includes(key)
            ? { ...value, [key]: Default(additionalProperties, value[key]) }
            : value;
    }, knownProperties);
}
// prettier-ignore
function TRecord(schema, references, value) {
    const defaulted = Default(schema, value);
    if (!IsPlainObject(value))
        return defaulted;
    const pattern = Object.getOwnPropertyNames(schema.patternProperties)[0];
    const knownKeys = new RegExp(pattern);
    const knownProperties = Object.getOwnPropertyNames(value).reduce((value, key) => {
        return knownKeys.test(key)
            ? { ...value, [key]: Visit(schema.patternProperties[pattern], references, value[key]) }
            : value;
    }, defaulted);
    if (!IsSchemaType(schema.additionalProperties)) {
        return Default(schema, knownProperties);
    }
    const unknownKeys = Object.getOwnPropertyNames(knownProperties);
    const additionalProperties = schema.additionalProperties;
    return unknownKeys.reduce((value, key) => {
        return !knownKeys.test(key)
            ? { ...value, [key]: Default(additionalProperties, value[key]) }
            : value;
    }, knownProperties);
}
// prettier-ignore
function TRef(schema, references, value) {
    const target = Deref(schema, references);
    const resolved = Visit(target, references, value);
    return Default(schema, resolved);
}
// prettier-ignore
function TThis(schema, references, value) {
    const target = Deref(schema, references);
    const resolved = Visit(target, references, value);
    return Default(schema, resolved);
}
// prettier-ignore
function TTuple(schema, references, value) {
    const value1 = Default(schema, value);
    return IsArray(schema.items) ? schema.items.map((schema, index) => Visit(schema, references, value1[index])) : [];
}
// prettier-ignore
function TUnion(schema, references, value) {
    // test value against union variants
    for (const subschema of schema.anyOf) {
        if (!Check(subschema, references, value))
            continue;
        const value1 = Visit(subschema, references, value);
        return Default(schema, value1);
    }
    // test transformed value against union variants
    for (const subschema of schema.anyOf) {
        const value1 = Visit(subschema, references, value);
        if (!Check(schema, references, value1))
            continue;
        return Default(schema, value1);
    }
    return Default(schema, value);
}
// prettier-ignore
function Visit(schema, references, value) {
    const references_ = typeof schema.$id === 'string' ? [...references, schema] : references;
    const schema_ = schema;
    switch (schema[Kind]) {
        case 'Array':
            return TArray(schema_, references_, value);
        case 'Intersect':
            return TIntersect(schema_, references_, value);
        case 'Not':
            return TNot(schema_, references_, value);
        case 'Object':
            return TObject(schema_, references_, value);
        case 'Record':
            return TRecord(schema_, references_, value);
        case 'Ref':
            return TRef(schema_, references_, value);
        case 'This':
            return TThis(schema_, references_, value);
        case 'Tuple':
            return TTuple(schema_, references_, value);
        case 'Union':
            return TUnion(schema_, references_, value);
        default:
            return Default(schema_, value);
    }
}
/**
 * `[Internal]` Encodes the value and returns the result. This function expects the
 * caller to pass a statically checked value. This function does not check the encoded
 * result, meaning the result should be passed to `Check` before use. Refer to the
 * `Value.Encode()` function for implementation details.
 */
export function TransformEncode(schema, references, value) {
    return Visit(schema, references, value);
}
