/*--------------------------------------------------------------------------

@sinclair/typebox/value

The MIT License (MIT)

Copyright (c) 2017-2023 Haydn Paterson (sinclair) <haydn.developer@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

---------------------------------------------------------------------------*/
import { TTransform as IsTransformType, TSchema as IsSchemaType } from '../../type/guard/type.mjs';
import { IsString, IsUndefined } from '../guard/index.mjs';
import { Deref } from '../deref/index.mjs';
import { Kind } from '../../type/symbols/index.mjs';
// prettier-ignore
function TArray(schema, references) {
    return IsTransformType(schema) || Visit(schema.items, references);
}
// prettier-ignore
function TAsyncIterator(schema, references) {
    return IsTransformType(schema) || Visit(schema.items, references);
}
// prettier-ignore
function TConstructor(schema, references) {
    return IsTransformType(schema) || Visit(schema.returns, references) || schema.parameters.some((schema) => Visit(schema, references));
}
// prettier-ignore
function TFunction(schema, references) {
    return IsTransformType(schema) || Visit(schema.returns, references) || schema.parameters.some((schema) => Visit(schema, references));
}
// prettier-ignore
function TIntersect(schema, references) {
    return IsTransformType(schema) || IsTransformType(schema.unevaluatedProperties) || schema.allOf.some((schema) => Visit(schema, references));
}
// prettier-ignore
function TIterator(schema, references) {
    return IsTransformType(schema) || Visit(schema.items, references);
}
// prettier-ignore
function TNot(schema, references) {
    return IsTransformType(schema) || Visit(schema.not, references);
}
// prettier-ignore
function TObject(schema, references) {
    return (IsTransformType(schema) ||
        Object.values(schema.properties).some((schema) => Visit(schema, references)) ||
        (IsSchemaType(schema.additionalProperties) && Visit(schema.additionalProperties, references)));
}
// prettier-ignore
function TPromise(schema, references) {
    return IsTransformType(schema) || Visit(schema.item, references);
}
// prettier-ignore
function TRecord(schema, references) {
    const pattern = Object.getOwnPropertyNames(schema.patternProperties)[0];
    const property = schema.patternProperties[pattern];
    return IsTransformType(schema) || Visit(property, references) || (IsSchemaType(schema.additionalProperties) && IsTransformType(schema.additionalProperties));
}
// prettier-ignore
function TRef(schema, references) {
    if (IsTransformType(schema))
        return true;
    return Visit(Deref(schema, references), references);
}
// prettier-ignore
function TThis(schema, references) {
    if (IsTransformType(schema))
        return true;
    return Visit(Deref(schema, references), references);
}
// prettier-ignore
function TTuple(schema, references) {
    return IsTransformType(schema) || (!IsUndefined(schema.items) && schema.items.some((schema) => Visit(schema, references)));
}
// prettier-ignore
function TUnion(schema, references) {
    return IsTransformType(schema) || schema.anyOf.some((schema) => Visit(schema, references));
}
// prettier-ignore
function Visit(schema, references) {
    const references_ = IsString(schema.$id) ? [...references, schema] : references;
    const schema_ = schema;
    if (schema.$id && visited.has(schema.$id))
        return false;
    if (schema.$id)
        visited.add(schema.$id);
    switch (schema[Kind]) {
        case 'Array':
            return TArray(schema_, references_);
        case 'AsyncIterator':
            return TAsyncIterator(schema_, references_);
        case 'Constructor':
            return TConstructor(schema_, references_);
        case 'Function':
            return TFunction(schema_, references_);
        case 'Intersect':
            return TIntersect(schema_, references_);
        case 'Iterator':
            return TIterator(schema_, references_);
        case 'Not':
            return TNot(schema_, references_);
        case 'Object':
            return TObject(schema_, references_);
        case 'Promise':
            return TPromise(schema_, references_);
        case 'Record':
            return TRecord(schema_, references_);
        case 'Ref':
            return TRef(schema_, references_);
        case 'This':
            return TThis(schema_, references_);
        case 'Tuple':
            return TTuple(schema_, references_);
        case 'Union':
            return TUnion(schema_, references_);
        default:
            return IsTransformType(schema);
    }
}
const visited = new Set();
/** Returns true if this schema contains a transform codec */
export function HasTransform(schema, references) {
    visited.clear();
    return Visit(schema, references);
}
