"use strict";
/*--------------------------------------------------------------------------

@sinclair/typebox/type

The MIT License (MIT)

Copyright (c) 2017-2023 Haydn Paterson (sinclair) <haydn.developer@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

---------------------------------------------------------------------------*/
Object.defineProperty(exports, "__esModule", { value: true });
exports.TSchema = exports.TKind = exports.TVoid = exports.TUnsafe = exports.TUnknown = exports.TUint8Array = exports.TUnion = exports.TUnionLiteral = exports.TUndefined = exports.TTuple = exports.TTransform = exports.TThis = exports.TTemplateLiteral = exports.TSymbol = exports.TString = exports.TRef = exports.TRecursive = exports.TRecord = exports.TPromise = exports.TObject = exports.TNumber = exports.TNull = exports.TNot = exports.TNever = exports.TMappedResult = exports.TMappedKey = exports.TLiteralValue = exports.TLiteral = exports.TLiteralBoolean = exports.TLiteralNumber = exports.TLiteralString = exports.TKindOf = exports.TIterator = exports.TIntersect = exports.TProperties = exports.TInteger = exports.TFunction = exports.TDate = exports.TConstructor = exports.TBoolean = exports.TBigInt = exports.TAsyncIterator = exports.TArray = exports.TAny = exports.TOptional = exports.TReadonly = exports.TypeGuardUnknownTypeError = void 0;
const ValueGuard = require("./value");
const index_1 = require("../symbols/index");
class TypeGuardUnknownTypeError extends Error {
}
exports.TypeGuardUnknownTypeError = TypeGuardUnknownTypeError;
const KnownTypes = [
    'Any',
    'Array',
    'AsyncIterator',
    'BigInt',
    'Boolean',
    'Constructor',
    'Date',
    'Enum',
    'Function',
    'Integer',
    'Intersect',
    'Iterator',
    'Literal',
    'MappedKey',
    'MappedResult',
    'Not',
    'Null',
    'Number',
    'Object',
    'Promise',
    'Record',
    'Ref',
    'String',
    'Symbol',
    'TemplateLiteral',
    'This',
    'Tuple',
    'Undefined',
    'Union',
    'Uint8Array',
    'Unknown',
    'Void',
];
function IsPattern(value) {
    try {
        new RegExp(value);
        return true;
    }
    catch {
        return false;
    }
}
function IsControlCharacterFree(value) {
    if (!ValueGuard.IsString(value))
        return false;
    for (let i = 0; i < value.length; i++) {
        const code = value.charCodeAt(i);
        if ((code >= 7 && code <= 13) || code === 27 || code === 127) {
            return false;
        }
    }
    return true;
}
function IsAdditionalProperties(value) {
    return IsOptionalBoolean(value) || (0, index_2.TSchema)(value);
}
function IsOptionalBigInt(value) {
    return ValueGuard.IsUndefined(value) || ValueGuard.IsBigInt(value);
}
function IsOptionalNumber(value) {
    return ValueGuard.IsUndefined(value) || ValueGuard.IsNumber(value);
}
function IsOptionalBoolean(value) {
    return ValueGuard.IsUndefined(value) || ValueGuard.IsBoolean(value);
}
function IsOptionalString(value) {
    return ValueGuard.IsUndefined(value) || ValueGuard.IsString(value);
}
function IsOptionalPattern(value) {
    return ValueGuard.IsUndefined(value) || (ValueGuard.IsString(value) && IsControlCharacterFree(value) && IsPattern(value));
}
function IsOptionalFormat(value) {
    return ValueGuard.IsUndefined(value) || (ValueGuard.IsString(value) && IsControlCharacterFree(value));
}
function IsOptionalSchema(value) {
    return ValueGuard.IsUndefined(value) || (0, index_2.TSchema)(value);
}
// ------------------------------------------------------------------
// Modifiers
// ------------------------------------------------------------------
/** Returns true if this value has a Readonly symbol */
function TReadonly(value) {
    return ValueGuard.IsObject(value) && value[index_1.ReadonlyKind] === 'Readonly';
}
exports.TReadonly = TReadonly;
/** Returns true if this value has a Optional symbol */
function TOptional(value) {
    return ValueGuard.IsObject(value) && value[index_1.OptionalKind] === 'Optional';
}
exports.TOptional = TOptional;
// ------------------------------------------------------------------
// Types
// ------------------------------------------------------------------
/** Returns true if the given value is TAny */
function TAny(value) {
    // prettier-ignore
    return (TKindOf(value, 'Any') &&
        IsOptionalString(value.$id));
}
exports.TAny = TAny;
/** Returns true if the given value is TArray */
function TArray(value) {
    return (TKindOf(value, 'Array') &&
        value.type === 'array' &&
        IsOptionalString(value.$id) &&
        (0, index_2.TSchema)(value.items) &&
        IsOptionalNumber(value.minItems) &&
        IsOptionalNumber(value.maxItems) &&
        IsOptionalBoolean(value.uniqueItems) &&
        IsOptionalSchema(value.contains) &&
        IsOptionalNumber(value.minContains) &&
        IsOptionalNumber(value.maxContains));
}
exports.TArray = TArray;
/** Returns true if the given value is TAsyncIterator */
function TAsyncIterator(value) {
    // prettier-ignore
    return (TKindOf(value, 'AsyncIterator') &&
        value.type === 'AsyncIterator' &&
        IsOptionalString(value.$id) &&
        (0, index_2.TSchema)(value.items));
}
exports.TAsyncIterator = TAsyncIterator;
/** Returns true if the given value is TBigInt */
function TBigInt(value) {
    // prettier-ignore
    return (TKindOf(value, 'BigInt') &&
        value.type === 'bigint' &&
        IsOptionalString(value.$id) &&
        IsOptionalBigInt(value.exclusiveMaximum) &&
        IsOptionalBigInt(value.exclusiveMinimum) &&
        IsOptionalBigInt(value.maximum) &&
        IsOptionalBigInt(value.minimum) &&
        IsOptionalBigInt(value.multipleOf));
}
exports.TBigInt = TBigInt;
/** Returns true if the given value is TBoolean */
function TBoolean(value) {
    // prettier-ignore
    return (TKindOf(value, 'Boolean') &&
        value.type === 'boolean' &&
        IsOptionalString(value.$id));
}
exports.TBoolean = TBoolean;
/** Returns true if the given value is TConstructor */
function TConstructor(value) {
    // prettier-ignore
    return (TKindOf(value, 'Constructor') &&
        value.type === 'Constructor' &&
        IsOptionalString(value.$id) &&
        ValueGuard.IsArray(value.parameters) &&
        value.parameters.every(schema => (0, index_2.TSchema)(schema)) &&
        (0, index_2.TSchema)(value.returns));
}
exports.TConstructor = TConstructor;
/** Returns true if the given value is TDate */
function TDate(value) {
    return (TKindOf(value, 'Date') &&
        value.type === 'Date' &&
        IsOptionalString(value.$id) &&
        IsOptionalNumber(value.exclusiveMaximumTimestamp) &&
        IsOptionalNumber(value.exclusiveMinimumTimestamp) &&
        IsOptionalNumber(value.maximumTimestamp) &&
        IsOptionalNumber(value.minimumTimestamp) &&
        IsOptionalNumber(value.multipleOfTimestamp));
}
exports.TDate = TDate;
/** Returns true if the given value is TFunction */
function TFunction(value) {
    // prettier-ignore
    return (TKindOf(value, 'Function') &&
        value.type === 'Function' &&
        IsOptionalString(value.$id) &&
        ValueGuard.IsArray(value.parameters) &&
        value.parameters.every(schema => (0, index_2.TSchema)(schema)) &&
        (0, index_2.TSchema)(value.returns));
}
exports.TFunction = TFunction;
/** Returns true if the given value is TInteger */
function TInteger(value) {
    return (TKindOf(value, 'Integer') &&
        value.type === 'integer' &&
        IsOptionalString(value.$id) &&
        IsOptionalNumber(value.exclusiveMaximum) &&
        IsOptionalNumber(value.exclusiveMinimum) &&
        IsOptionalNumber(value.maximum) &&
        IsOptionalNumber(value.minimum) &&
        IsOptionalNumber(value.multipleOf));
}
exports.TInteger = TInteger;
/** Returns true if the given schema is TProperties */
function TProperties(value) {
    // prettier-ignore
    return (ValueGuard.IsObject(value) &&
        Object.entries(value).every(([key, schema]) => IsControlCharacterFree(key) && (0, index_2.TSchema)(schema)));
}
exports.TProperties = TProperties;
/** Returns true if the given value is TIntersect */
function TIntersect(value) {
    // prettier-ignore
    return (TKindOf(value, 'Intersect') &&
        (ValueGuard.IsString(value.type) && value.type !== 'object' ? false : true) &&
        ValueGuard.IsArray(value.allOf) &&
        value.allOf.every(schema => (0, index_2.TSchema)(schema) && !TTransform(schema)) &&
        IsOptionalString(value.type) &&
        (IsOptionalBoolean(value.unevaluatedProperties) || IsOptionalSchema(value.unevaluatedProperties)) &&
        IsOptionalString(value.$id));
}
exports.TIntersect = TIntersect;
/** Returns true if the given value is TIterator */
function TIterator(value) {
    // prettier-ignore
    return (TKindOf(value, 'Iterator') &&
        value.type === 'Iterator' &&
        IsOptionalString(value.$id) &&
        (0, index_2.TSchema)(value.items));
}
exports.TIterator = TIterator;
/** Returns true if the given value is a TKind with the given name. */
function TKindOf(value, kind) {
    return ValueGuard.IsObject(value) && index_1.Kind in value && value[index_1.Kind] === kind;
}
exports.TKindOf = TKindOf;
/** Returns true if the given value is TLiteral<string> */
function TLiteralString(value) {
    return (0, index_3.TLiteral)(value) && ValueGuard.IsString(value.const);
}
exports.TLiteralString = TLiteralString;
/** Returns true if the given value is TLiteral<number> */
function TLiteralNumber(value) {
    return (0, index_3.TLiteral)(value) && ValueGuard.IsNumber(value.const);
}
exports.TLiteralNumber = TLiteralNumber;
/** Returns true if the given value is TLiteral<boolean> */
function TLiteralBoolean(value) {
    return (0, index_3.TLiteral)(value) && ValueGuard.IsBoolean(value.const);
}
exports.TLiteralBoolean = TLiteralBoolean;
/** Returns true if the given value is TLiteral */
function TLiteral(value) {
    // prettier-ignore
    return (TKindOf(value, 'Literal') &&
        IsOptionalString(value.$id) && (0, index_3.TLiteralValue)(value.const));
}
exports.TLiteral = TLiteral;
/** Returns true if the given value is a TLiteralValue */
function TLiteralValue(value) {
    return ValueGuard.IsBoolean(value) || ValueGuard.IsNumber(value) || ValueGuard.IsString(value);
}
exports.TLiteralValue = TLiteralValue;
/** Returns true if the given value is a TMappedKey */
function TMappedKey(value) {
    // prettier-ignore
    return (TKindOf(value, 'MappedKey') &&
        ValueGuard.IsArray(value.keys) &&
        value.keys.every(key => ValueGuard.IsNumber(key) || ValueGuard.IsString(key)));
}
exports.TMappedKey = TMappedKey;
/** Returns true if the given value is TMappedResult */
function TMappedResult(value) {
    // prettier-ignore
    return (TKindOf(value, 'MappedResult') &&
        (0, index_4.TProperties)(value.properties));
}
exports.TMappedResult = TMappedResult;
/** Returns true if the given value is TNever */
function TNever(value) {
    // prettier-ignore
    return (TKindOf(value, 'Never') &&
        ValueGuard.IsObject(value.not) &&
        Object.getOwnPropertyNames(value.not).length === 0);
}
exports.TNever = TNever;
/** Returns true if the given value is TNot */
function TNot(value) {
    // prettier-ignore
    return (TKindOf(value, 'Not') &&
        (0, index_2.TSchema)(value.not));
}
exports.TNot = TNot;
/** Returns true if the given value is TNull */
function TNull(value) {
    // prettier-ignore
    return (TKindOf(value, 'Null') &&
        value.type === 'null' &&
        IsOptionalString(value.$id));
}
exports.TNull = TNull;
/** Returns true if the given value is TNumber */
function TNumber(value) {
    return (TKindOf(value, 'Number') &&
        value.type === 'number' &&
        IsOptionalString(value.$id) &&
        IsOptionalNumber(value.exclusiveMaximum) &&
        IsOptionalNumber(value.exclusiveMinimum) &&
        IsOptionalNumber(value.maximum) &&
        IsOptionalNumber(value.minimum) &&
        IsOptionalNumber(value.multipleOf));
}
exports.TNumber = TNumber;
/** Returns true if the given value is TObject */
function TObject(value) {
    // prettier-ignore
    return (TKindOf(value, 'Object') &&
        value.type === 'object' &&
        IsOptionalString(value.$id) &&
        (0, index_4.TProperties)(value.properties) &&
        IsAdditionalProperties(value.additionalProperties) &&
        IsOptionalNumber(value.minProperties) &&
        IsOptionalNumber(value.maxProperties));
}
exports.TObject = TObject;
/** Returns true if the given value is TPromise */
function TPromise(value) {
    // prettier-ignore
    return (TKindOf(value, 'Promise') &&
        value.type === 'Promise' &&
        IsOptionalString(value.$id) &&
        (0, index_2.TSchema)(value.item));
}
exports.TPromise = TPromise;
/** Returns true if the given value is TRecord */
function TRecord(value) {
    // prettier-ignore
    return (TKindOf(value, 'Record') &&
        value.type === 'object' &&
        IsOptionalString(value.$id) &&
        IsAdditionalProperties(value.additionalProperties) &&
        ValueGuard.IsObject(value.patternProperties) &&
        ((schema) => {
            const keys = Object.getOwnPropertyNames(schema.patternProperties);
            return (keys.length === 1 &&
                IsPattern(keys[0]) &&
                ValueGuard.IsObject(schema.patternProperties) &&
                (0, index_2.TSchema)(schema.patternProperties[keys[0]]));
        })(value));
}
exports.TRecord = TRecord;
/** Returns true if this value is TRecursive */
function TRecursive(value) {
    return ValueGuard.IsObject(value) && index_1.Hint in value && value[index_1.Hint] === 'Recursive';
}
exports.TRecursive = TRecursive;
/** Returns true if the given value is TRef */
function TRef(value) {
    // prettier-ignore
    return (TKindOf(value, 'Ref') &&
        IsOptionalString(value.$id) &&
        ValueGuard.IsString(value.$ref));
}
exports.TRef = TRef;
/** Returns true if the given value is TString */
function TString(value) {
    // prettier-ignore
    return (TKindOf(value, 'String') &&
        value.type === 'string' &&
        IsOptionalString(value.$id) &&
        IsOptionalNumber(value.minLength) &&
        IsOptionalNumber(value.maxLength) &&
        IsOptionalPattern(value.pattern) &&
        IsOptionalFormat(value.format));
}
exports.TString = TString;
/** Returns true if the given value is TSymbol */
function TSymbol(value) {
    // prettier-ignore
    return (TKindOf(value, 'Symbol') &&
        value.type === 'symbol' &&
        IsOptionalString(value.$id));
}
exports.TSymbol = TSymbol;
/** Returns true if the given value is TTemplateLiteral */
function TTemplateLiteral(value) {
    // prettier-ignore
    return (TKindOf(value, 'TemplateLiteral') &&
        value.type === 'string' &&
        ValueGuard.IsString(value.pattern) &&
        value.pattern[0] === '^' &&
        value.pattern[value.pattern.length - 1] === '$');
}
exports.TTemplateLiteral = TTemplateLiteral;
/** Returns true if the given value is TThis */
function TThis(value) {
    // prettier-ignore
    return (TKindOf(value, 'This') &&
        IsOptionalString(value.$id) &&
        ValueGuard.IsString(value.$ref));
}
exports.TThis = TThis;
/** Returns true of this value is TTransform */
function TTransform(value) {
    return ValueGuard.IsObject(value) && index_1.TransformKind in value;
}
exports.TTransform = TTransform;
/** Returns true if the given value is TTuple */
function TTuple(value) {
    // prettier-ignore
    return (TKindOf(value, 'Tuple') &&
        value.type === 'array' &&
        IsOptionalString(value.$id) &&
        ValueGuard.IsNumber(value.minItems) &&
        ValueGuard.IsNumber(value.maxItems) &&
        value.minItems === value.maxItems &&
        (( // empty
        ValueGuard.IsUndefined(value.items) &&
            ValueGuard.IsUndefined(value.additionalItems) &&
            value.minItems === 0) || (ValueGuard.IsArray(value.items) &&
            value.items.every(schema => (0, index_2.TSchema)(schema)))));
}
exports.TTuple = TTuple;
/** Returns true if the given value is TUndefined */
function TUndefined(value) {
    // prettier-ignore
    return (TKindOf(value, 'Undefined') &&
        value.type === 'undefined' &&
        IsOptionalString(value.$id));
}
exports.TUndefined = TUndefined;
/** Returns true if the given value is TUnion<Literal<string | number>[]> */
function TUnionLiteral(value) {
    return (0, index_5.TUnion)(value) && value.anyOf.every((schema) => TLiteralString(schema) || TLiteralNumber(schema));
}
exports.TUnionLiteral = TUnionLiteral;
/** Returns true if the given value is TUnion */
function TUnion(value) {
    // prettier-ignore
    return (TKindOf(value, 'Union') &&
        IsOptionalString(value.$id) &&
        ValueGuard.IsObject(value) &&
        ValueGuard.IsArray(value.anyOf) &&
        value.anyOf.every(schema => (0, index_2.TSchema)(schema)));
}
exports.TUnion = TUnion;
/** Returns true if the given value is TUint8Array */
function TUint8Array(value) {
    // prettier-ignore
    return (TKindOf(value, 'Uint8Array') &&
        value.type === 'Uint8Array' &&
        IsOptionalString(value.$id) &&
        IsOptionalNumber(value.minByteLength) &&
        IsOptionalNumber(value.maxByteLength));
}
exports.TUint8Array = TUint8Array;
/** Returns true if the given value is TUnknown */
function TUnknown(value) {
    // prettier-ignore
    return (TKindOf(value, 'Unknown') &&
        IsOptionalString(value.$id));
}
exports.TUnknown = TUnknown;
/** Returns true if the given value is a raw TUnsafe */
function TUnsafe(value) {
    return TKindOf(value, 'Unsafe');
}
exports.TUnsafe = TUnsafe;
/** Returns true if the given value is TVoid */
function TVoid(value) {
    // prettier-ignore
    return (TKindOf(value, 'Void') &&
        value.type === 'void' &&
        IsOptionalString(value.$id));
}
exports.TVoid = TVoid;
/** Returns true if the given value is TKind */
function TKind(value) {
    return ValueGuard.IsObject(value) && index_1.Kind in value && ValueGuard.IsString(value[index_1.Kind]) && !KnownTypes.includes(value[index_1.Kind]);
}
exports.TKind = TKind;
/** Returns true if the given value is TSchema */
function TSchema(value) {
    // prettier-ignore
    return (ValueGuard.IsObject(value)) && ((0, index_6.TAny)(value) ||
        (0, index_7.TArray)(value) ||
        (0, index_8.TBoolean)(value) ||
        (0, index_9.TBigInt)(value) ||
        (0, index_10.TAsyncIterator)(value) ||
        (0, index_11.TConstructor)(value) ||
        (0, index_12.TDate)(value) ||
        (0, index_13.TFunction)(value) ||
        (0, index_14.TInteger)(value) ||
        (0, index_15.TIntersect)(value) ||
        (0, index_16.TIterator)(value) ||
        (0, index_3.TLiteral)(value) ||
        (0, index_17.TMappedKey)(value) ||
        (0, index_17.TMappedResult)(value) ||
        (0, index_18.TNever)(value) ||
        (0, index_19.TNot)(value) ||
        (0, index_20.TNull)(value) ||
        (0, index_21.TNumber)(value) ||
        (0, index_4.TObject)(value) ||
        (0, index_22.TPromise)(value) ||
        (0, index_23.TRecord)(value) ||
        (0, index_24.TRef)(value) ||
        (0, index_25.TString)(value) ||
        (0, index_26.TSymbol)(value) ||
        (0, index_27.TTemplateLiteral)(value) ||
        (0, index_28.TThis)(value) ||
        (0, index_29.TTuple)(value) ||
        (0, index_30.TUndefined)(value) ||
        (0, index_5.TUnion)(value) ||
        (0, index_31.TUint8Array)(value) ||
        (0, index_32.TUnknown)(value) ||
        (0, index_33.TUnsafe)(value) ||
        (0, index_34.TVoid)(value) ||
        TKind(value));
}
exports.TSchema = TSchema;
